# Copyright 2020 Andrzej Cichocki

# This file is part of soak.
#
# soak is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# soak is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with soak.  If not, see <http://www.gnu.org/licenses/>.

from .util import PathResolvable, Snapshot
from aridity import NoSuchPathException, Repl
from aridity.model import Directive, Function, Text
from aridity.scope import Scope, slashfunction
from lagoon import git
from lagoon.program import ONELINE
from pathlib import Path
import os, re, subprocess, yaml

singledigit = re.compile('[0-9]')
zeroormorespaces = re.compile(' *')
zeroormoredots = re.compile('[.]*')
linefeed = '\n'
dotpy = '.py'
toplevelres = PathResolvable('toplevel')

def plugin(prefix, suffix, scope):
    modulename, globalname = (obj.cat() for _, obj in suffix.tophrase().resolve(scope, aslist = True).itero())
    leadingdots = len(zeroormoredots.match(modulename).group())
    words = modulename[leadingdots:].split('.')
    relpath = Path(*words[:-1]) / f"{words[-1]}{dotpy}"
    if leadingdots:
        modulepath = Path(scope.resolved('here').cat(), *['..'] * (leadingdots - 1), relpath)
        try:
            modulename = str(modulepath.relative_to(toplevelres.resolve(scope).cat()))[:-len(dotpy)].replace(os.sep, '.')
        except NoSuchPathException:
            modulename = None # It won't be able to do its own relative imports.
    else:
        modulepath = Path(toplevelres.resolve(scope).cat(), relpath)
    g = {} if modulename is None else dict(__name__ = modulename)
    with modulepath.open() as f:
        exec(f.read(), g)
    g[globalname](scope)

def blockliteral(scope, textresolvable):
    text = yaml.dump(textresolvable.resolve(scope).cat(), default_style = '|')
    header, *lines = text.splitlines() # For template interpolation convenience we discard the (insignificant) trailing newline.
    if not lines:
        return Text(header)
    if '...' == lines[-1]:
        lines.pop() # XXX: Could this result in no remaining lines?
    indentunit = scope.resolved('indentunit').cat()
    m = singledigit.search(header)
    if m is None:
        pyyamlindent = len(zeroormorespaces.match(lines[0]).group())
    else:
        pyyamlindent = int(m.group())
        header = f"{header[:m.start()]}{len(zeroormorespaces.fullmatch(indentunit).group())}{header[m.end():]}"
    contextindent = scope.resolved('indent').cat()
    return Text(f"""{header}\n{linefeed.join(f"{contextindent}{indentunit}{line[pyyamlindent:]}" for line in lines)}""")

def rootpath(scope, *resolvables):
    return slashfunction(scope, toplevelres, *resolvables)

def _toplevel(anydir):
    try:
        return Text(git.rev_parse.__show_toplevel[ONELINE](cwd = anydir))
    except subprocess.CalledProcessError:
        raise NoSuchPathException('Git property: toplevel')

def createparent(soakroot):
    parent = Scope()
    parent['plugin',] = Directive(plugin)
    parent['|',] = Function(blockliteral)
    parent['//',] = Function(rootpath)
    parent['toplevel',] = Snapshot(lambda: _toplevel(soakroot))
    with Repl(parent) as repl:
        repl('data = $processtemplate$(from)') # XXX: Too easy to accidentally override?
        repl.printf("indentunit = %s", 4 * ' ')
    return parent



# Copyright 2020 Andrzej Cichocki

# This file is part of soak.
#
# soak is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# soak is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with soak.  If not, see <http://www.gnu.org/licenses/>.

from . import cpuexecutor
from .context import createparent
from .terminal import LogFile, Terminal
from argparse import ArgumentParser
from aridity import Repl
from diapyr.util import invokeall
from functools import partial
from lagoon import diff
from pathlib import Path
import logging, os

class SoakConfig:

    soakkey = 'soak'

    def __init__(self, parent, configpath):
        self.scope = parent.createchild()
        with Repl(self.scope) as repl:
            repl.printf("cwd = %s", configpath.parent.resolve())
            repl.printf(". %s", configpath.name)
        self.reltargets = [Path(rt) for rt, _ in self.scope.resolved(self.soakkey).resolvables.items()]
        self.dirpath = configpath.parent

    def process(self, log, reltarget):
        relpartial = reltarget.with_name(f"{reltarget.name}.part")
        target = self.dirpath / reltarget
        log(target, rev = True)
        self.scope.resolved(self.soakkey, str(reltarget), 'data').writeout(self.dirpath / relpartial)
        (self.dirpath / relpartial).rename(target)
        log(target)

    def origtext(self, reltarget):
        return self.scope.resolved(self.soakkey, str(reltarget), 'diff').cat()

    def diff(self, origtext, reltarget):
        diff._us[print]('--color=always', '-', self.dirpath / reltarget, input = origtext, check = False)

def main_soak():
    'Process aridity templates as per all soak.arid configs in directory tree.'
    parser = ArgumentParser()
    parser.add_argument('-n', action = 'store_true')
    parser.add_argument('-d', action = 'store_true')
    parser.add_argument('-v', action = 'store_true')
    soak(parser.parse_args(), Path('.'))

def soak(config, soakroot):
    logging.basicConfig(format = "[%(levelname)s] %(message)s", level = logging.DEBUG if config.v else logging.INFO)
    parent = createparent(soakroot)
    soakconfigs = [SoakConfig(parent, p) for p in soakroot.rglob('soak.arid')]
    if not config.n:
        terminal = Terminal(sum(len(sc.reltargets) for sc in soakconfigs)) if 'TERM' in os.environ else LogFile
        with cpuexecutor() as executor:
            results = []
            for soakconfig in soakconfigs:
                for reltarget in soakconfig.reltargets:
                    log = partial(terminal.log, len(results))
                    log(soakconfig.dirpath / reltarget, dark = True)
                    results.append(executor.submit(soakconfig.process, log, reltarget).result)
            invokeall(results)
    if config.d:
        with cpuexecutor() as executor:
            diffs = []
            for soakconfig in soakconfigs:
                for reltarget in soakconfig.reltargets:
                    diffs.append(partial(lambda soakconfig, origtextfuture, reltarget: soakconfig.diff(origtextfuture.result(), reltarget),
                            soakconfig, executor.submit(soakconfig.origtext, reltarget), reltarget))
            invokeall(diffs)



# Copyright 2020 Andrzej Cichocki

# This file is part of soak.
#
# soak is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# soak is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with soak.  If not, see <http://www.gnu.org/licenses/>.

from diapyr.util import singleton
from lagoon import tput
import sys

class AbstractLog:

    stream = sys.stderr

    def log(self, index, obj, rev = False, dark = False):
        return self.logimpl(index, obj, rev, dark)

class Terminal(AbstractLog):

    def __init__(self, height):
        self.stream.write('\n' * height)
        self.height = height

    def logimpl(self, index, obj, rev, dark):
        def g():
            dy = self.height - index
            yield tput.cuu(dy)
            if rev:
                yield tput.rev()
            if dark:
                yield tput.setaf(0)
            yield str(obj)
            yield '\r'
            yield tput.sgr0()
            yield tput.cud(dy)
        self.stream.write(''.join(g()))
        self.stream.flush()

@singleton
class LogFile(AbstractLog):

    def logimpl(self, index, obj, rev, dark):
        if not dark:
            print('Damp:' if rev else 'Soaked:', obj, file = self.stream)



# Copyright 2020 Andrzej Cichocki

# This file is part of soak.
#
# soak is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# soak is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with soak.  If not, see <http://www.gnu.org/licenses/>.

from aridity.model import Resolvable
from threading import Lock

class Snapshot(Resolvable):

    class Result:

        def __init__(self, eornone, obj):
            self.eornone = eornone
            self.obj = obj

        def get(self):
            if self.eornone is not None:
                raise self.eornone
            return self.obj

    def __init__(self, factory):
        self.lock = Lock()
        self.factory = factory

    def _loadresult(self):
        with self.lock:
            try:
                self.result
            except AttributeError:
                f = self.factory
                try:
                    obj = f()
                    eornone = None
                except Exception as e:
                    obj = None
                    eornone = e
                self.result = self.Result(eornone, obj)

    def resolve(self, scope):
        try:
            r = self.result
        except AttributeError:
            self._loadresult()
            r = self.result
        return r.get()

class PathResolvable(Resolvable):

    def __init__(self, *path):
        self.path = path

    def resolve(self, scope):
        return scope.resolved(*self.path)
